@@ -1,4 +1,3 @@
-import Control.Monad (when)
 import qualified Data.Aeson as J
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
@@ -9,8 +8,9 @@ import qualified Data.Yaml as Y
 main :: IO ()
 main = do
     args <- getArgs
-    when (length args > 2) $ error "Usage: json2yaml [in] [out]"
-    let (input:output:_) = args ++ repeat "-"
+    (input, output) <- case args ++ replicate (2 - length args) "-" of
+        [i, o] -> return (i, o)
+        _ -> fail "Usage: json2yaml [in] [out]"
     mval <- fmap J.decode $
         case input of
             "-" -> L.getContents
